/*
 * C compiler
 * ==========
 *
 * Copyright 1989, 1990, 1991 Christoph van Wuellen.
 * Credits to Matthew Brandt.
 * All commercial rights reserved.
 *
 * This compiler may be redistributed as long there is no
 * commercial interest. The compiler must not be redistributed
 * without its full sources. This notice must stay intact.
 *
 * History:
 *
 * 1989   starting an 68000 C compiler, starting with material
 *        originally by M. Brandt
 * 1990   68000 C compiler further bug fixes
 *        started i386 port (December)
 * 1991   i386 port finished (January)
 *        further corrections in the front end and in the 68000
 *        code generator.
 *        The next port will be a SPARC port
 */

#include "config.h"

#ifdef INTEL

#include "chdr.h"
#include "expr.h"
#include "cglbdec.h"
#include "proto.h"
#include "genx86.h"
#include "outproto.h"

/*
 * Register allocation (for the expression evaluation)
 * This modules handles the management of scratch registers.
 * It keeps track of the allocated registers and of the stack
 * Although large parts are identical to the 68000 version,
 * the diffs are so big that I maintain two different files
 */

#if defined(__STDC__) || defined(__cplusplus)
#define P_(s) s
#else
#define P_(s) ()
#endif

static	void	g_push	P_((REG, DEEP));
static	void	g_pop	P_((REG, DEEP));

#undef P_

static	DEEP	reg_in_use[NUM_REGS];
static	REG	next_dreg;		/* next temporary register */
static	REG	next_areg;		/* next temporary register */
	REG	max_dreg = MAX_REG;	/* maximum temporary register */
	REG	max_areg = EDI;	/* maximum temporary register */

#define	MAX_REG_STACK	((DEEP)30)

static struct {
    REG		reg;
    DEEP	depth;
}               reg_stack[(int)MAX_REG_STACK + 1];
static DEEP	reg_stack_ptr;

static struct {
    REG		reg;
    BOOL	pushed;
}               reg_alloc[(int)MAX_REG_STACK + 1];
static DEEP	reg_depth;

/*
**	Define the registers which can be used to pass parameters.
*/
static REG	parameter_registers[] = {
		    EAX, EBX, ECX, EDX, ESI, EDI
};
static REGLIST parameter_list = {
		    (int)(sizeof (parameter_registers) / sizeof (REG)),
		    &parameter_registers[0]
};

/*
**	Define the registers which must be saved by a function if
**	they are used.
**	If the register is used to return a value then it needn't be
**	saved.
**	If the register is used to pass a parameter then it needn't be
**	saved.
*/
static	REG	saved_registers[] = {
		    EAX, EBX, ECX, EDX, ESI, EDI
};
static	REGLIST	saved_list = {
		    (int) (sizeof (saved_registers) / sizeof (REG)),
		    &saved_registers[0]
};

/*
**	Define the registers which are used to return the results
**	from a function call.
*/
static	REG	result_registers[] = {
		    EAX, EDX
};
static	REGLIST	result_list = {
		    (int) (sizeof (result_registers) / sizeof (REG)),
		    &result_registers[0]
};

static REGUSAGE	rusage = {
		    &parameter_list,
		    &saved_list,
		    &result_list
};
REGUSAGE *	reg_usage = &rusage;

/*
 * this routine generates code to push a register onto the stack
 */
static void
g_push P2(REG, reg, DEEP, number)
{
    ADDRESS   *ap;
    sync_stack();
    ap = mk_dreg(reg);
    g_code(op_push, small_option ? IL2:IL4, ap, NIL_ADDRESS);
    reg_stack[reg_stack_ptr].reg = reg;
    reg_stack[reg_stack_ptr].depth = number;

    /* is already pushed */
    if (reg_alloc[number].pushed)
	FATAL ((__FILE__,"g_push","reg %d already pushed", (int)reg));
    reg_alloc[number].pushed = TRUE;

    /* check on stack overflow */
    if (++reg_stack_ptr > MAX_REG_STACK)
	FATAL ((__FILE__,"g_push","register stack overflow"));
}

/*
 * generate code to pop a register from the stack.
 */
static void
g_pop P2(REG, reg, DEEP, depth)
{
    ADDRESS   *ap;

    /* check on stack underflow */
    if (reg_stack_ptr-- == EMPTY)
	FATAL ((__FILE__,"g_pop","register %d stack empty", (int)reg));

    /* check if the desired register really is on stack */
    if (reg_stack[reg_stack_ptr].depth != depth)
	FATAL ((__FILE__,"g_pop","register %d order %d, %d", (int)reg, (int)reg_stack[reg_stack_ptr].depth, (int)depth));

    /* check if the register which is restored is really void */
    if (reg_in_use[reg] != UNUSED)
	FATAL ((__FILE__,"g_pop","register %d in use", (int)reg));

    reg_in_use[reg] = depth;
    sync_stack();
    ap = mk_dreg(reg);
    g_code(op_pop, small_option ? IL2:IL4, ap, NIL_ADDRESS);
    /* clear the push_flag */
    reg_alloc[depth].pushed = FALSE;
}

/*
 * this routine should be called before each expression is evaluated to make
 * sure the stack is balanced and all of the registers are marked free.
 * This is also a good place to free all 'pseudo' registers in the stack frame
 * by clearing act_scratch.
 */
void
initstack P0(void)
{
    REG		 reg;
    for (reg=EAX; reg<= EBP; reg++)
        reg_in_use[reg]= UNUSED;
    next_dreg = EAX;
    next_areg = ESI;
    reg_stack_ptr = EMPTY;
    reg_depth = EMPTY;
    act_scratch = 0;
}

/*
 * this routine checks if all allocated registers were freed correctly
 */
void
checkstack P0(void)
{
    REG reg;
    for (reg=EAX; reg<= max_dreg; reg++)
        if (reg_in_use[reg] != UNUSED)
            FATAL ((__FILE__,"checkstack","register %d in use", (int)reg));
    for (reg=ESI; reg<= EDI; reg++)
        if (reg_in_use[reg] != UNUSED)
            FATAL ((__FILE__,"checkstack","register %d in use", (int)reg));
    if (next_dreg != EAX)
        FATAL ((__FILE__,"checkstack","3 %d", (int) next_dreg));
    if (next_areg != ESI)
        FATAL ((__FILE__,"checkstack","4 %d", (int) next_areg));
    if (reg_stack_ptr != EMPTY)
        FATAL ((__FILE__,"checkstack","5"));
    if (reg_depth != EMPTY)
        FATAL ((__FILE__,"checkstack","6"));
}

/*
**	validate() will make sure that if a register within an address
**	mode has been pushed onto the stack that it is popped back at
**	this time.
*/
void
validate P1(const ADDRESS*, ap)
{
    char reg;
    switch (ap->mode) {
      case am_dreg:
        reg = ap->preg;
	if (reg <= max_dreg && reg_alloc[ap->deep].pushed) {
	    g_pop(reg, ap->deep);
	}
	break;
      case am_areg:
      case am_ind:
        reg = ap->preg;
	if (reg <= max_areg && reg_alloc[ap->deep].pushed) {
	    g_pop(reg, ap->deep);
	}
	break;
      case am_indx2:
        reg = ap->sreg;
	if (reg <= EDI && reg_alloc[ap->deep].pushed) {
	    g_pop(reg, ap->deep);
	}
        break;
      case am_indx:
        reg = ap->preg;
	if (reg <= EDI && reg_alloc[ap->deep].pushed) {
	    g_pop(reg, ap->deep);
	}
	break;
      case am_mreg:
	reg = ap->sreg;
	if (reg <= max_dreg && reg_alloc[ap->deep].pushed) {
	    g_pop(reg, (DEEP)((int)ap->deep+1));
	}
	reg = ap->preg;
	if (reg <= max_dreg && reg_alloc[ap->deep].pushed) {
	    g_pop(reg, ap->deep);
	}
	break;
      default:
        return;
    }
}


static REG
allocate_register P0(void)
{
    REG	reg = next_dreg;
    /*
    **	if the register is in use, push it to the stack
    */
    if (reg_in_use[reg] != UNUSED) {
        g_push ( reg, reg_in_use[reg] );
    }

    reg_in_use[reg] = reg_depth;
    reg_alloc[reg_depth].reg = reg;
    reg_alloc[reg_depth].pushed = FALSE;

    if (next_dreg++ == max_dreg)
	next_dreg = EAX;		/* wrap around */

    if (reg_depth++ >= MAX_REG_STACK)
        FATAL ((__FILE__,"allocate_register","register stack overflow"));
    return reg;
}

static REG
allocate_index P0(void)
{
    REG	reg = next_areg;
    /*
    **	if the register is in use, push it to the stack
    */
    if (reg_in_use[reg] != UNUSED)
        g_push ( reg, reg_in_use[reg] );

    reg_in_use[reg] = reg_depth;
    reg_alloc[reg_depth].reg = reg;
    reg_alloc[reg_depth].pushed = FALSE;

    if (next_areg++ == EDI)
	next_areg = ESI;		/* wrap around */

    if (reg_depth++ >= MAX_REG_STACK)
        FATAL ((__FILE__,"allocate_index","reigster stack overflow"));
    return reg;
}

static void
deallocate_register P1(REG, reg)
{
    DEEP	number;
    switch (reg) {
      case EAX:
      case EDX:
      case EBX:
      case ECX:
	if (reg > max_dreg)
	    return;
	if (next_dreg-- == EAX)
	    next_dreg = max_dreg;
	break;
      case ESI:
      case EDI:
	if (next_areg-- == ESI)
	    next_areg = EDI;
	break;
      default:
	CANNOT_REACH_HERE();
	break;
    }
    number = reg_in_use[reg];
    reg_in_use[reg] = UNUSED;

    /* we should only free the most recently allocated register */
    if (reg_depth-- == EMPTY)
	FATAL ((__FILE__,"deallocate_register","register %d stack empty", (int)reg));

    if (reg_depth != number)
	FATAL ((__FILE__,"deallocate_register","register %d stack order (%d,%d)", (int)reg, (int)reg_depth, (int)number));

    /* the just freed register should not be on stack */
    if (reg_alloc[number].pushed)
	FATAL ((__FILE__,"deallocate_register","register %d pushed", (int)reg));
}

/*
**	Allocate a register
*/
ADDRESS   *
data_register P0(void)
{
    REG		reg = allocate_register();
    ADDRESS *	ap = mk_dreg(reg);
    ap->deep = reg_in_use[reg];
    return ap;
}

/*
**	Allocate an index register
*/
ADDRESS   *
address_register P0(void)
{
    REG		reg = allocate_index();
    ADDRESS *	ap = mk_areg(reg);
    ap->deep = reg_in_use[reg];
    return ap;
}

/*
**	Allocates a register which is appropriate to the flags parameter.
*/
ADDRESS *
temp_register P1(FLAGS, flags)
{
    if (flags & F_DREG)
	return data_register();
    if (flags & F_AREG)
	return address_register();
    return NIL_ADDRESS;
}

/*
**	Allocate 2 registers
*/
ADDRESS *
mdata_register P0(void)
{
    REG		reg1 = allocate_register();
    REG		reg2 = allocate_register();
    ADDRESS *	ap = mk_mreg(reg1, reg2);
    ap->deep = reg_in_use[reg1];
    return ap;
}

/*
**	Returns TRUE is the specified register is not available at "no cost"
**	(no push).
*/
BOOL
is_register_used P1(REG, reg)
{
    return (reg_in_use[reg] != UNUSED);
}

/*
 * tells if an address mode uses a scratch register
 */
BOOL
uses_temp P1(const ADDRESS*, ap)
{
    if (ap == NIL_ADDRESS)
        FATAL( (__FILE__,"uses_temp",""));
    switch (ap->mode) {
      case am_dreg:
        return (ap->preg <= max_dreg);
      case am_ind:
      case am_indx:
      case am_indx2:
        return (ap->sreg <= max_areg);
      default:
        return FALSE;
    }
}

/*
** release any temporary registers used in an addressing mode.
*/
void
freeop P1(const ADDRESS*, ap)
{
    DEEP	number;
    REG		reg;
    if (ap == NIL_ADDRESS)
	/* This can happen freeing a NOVALUE result */
	return;
    switch (ap->mode) {
      case am_mreg:
      case am_indx2:
	deallocate_register (ap->sreg);
	/*FALLTHRU*/
      case am_dreg:
      case am_areg:
      case am_ind:
      case am_indx:
	reg = ap->preg;
	break;
      default:
	return;
    }

    switch (reg) {
      case EAX:
      case EDX:
      case EBX:
      case ECX:
	if (reg > max_dreg)
	    return;
	break;
      case ESI:
      case EDI:
	break;
      case ESP:
      case EBP:
	return;
      default:
	CANNOT_REACH_HERE();
	break;
    }
    number = reg_in_use[reg];
    deallocate_register (reg);

    /* some consistency checks */
    if (number != ap->deep)
	FATAL ((__FILE__,"freeop","(%d, %d, %d)", (int)reg, (int)number, (int)ap->deep));
}

/*
**	push any used temporary registers.
**
**	This is necessary across function calls.
**	The reason for this hacking is actually that temp_inv should dump
**	the registers in the correct order,
**	the least recently allocate register first.
**	the most recently allocated register last.
*/
void
temp_inv P0(void)
{
    DEEP	deep;

    for (deep = EMPTY; deep < reg_depth; deep++)
	if (!reg_alloc[deep].pushed) {
	    g_push(reg_alloc[deep].reg, deep);
	    /* mark the register void */
            reg_in_use[reg_alloc[deep].reg] = UNUSED;
	}
}

/*
**	Converts a list of registers into a register mask
*/
REGMASK
reglist_to_mask P1(const REGLIST *, rp)
{
    REGMASK	mask = (REGMASK)0;
    int		num;
    for (num = 0; num < rp->number; num++) {
	mask |= (REGMASK)(1 << (int)rp->reg[num]);
    }
    return mask;
}
#endif /* INTEL */
